from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="pytrendline",
    version="1.0.1",
    author="Eduardo Nunez",
    author_email="ed@ednunez.me",
    description="Detection for support and resistance trendlines on python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ednunezg/pytrendline",
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    platforms=['any'],
    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'examples']),
    include_package_data=True,
    install_requires=[
        'numpy>=1.18.5',
        'pandas>=1.0.4',
        'datetime>=4.3',
        'bokeh>=2.0.2',
        'colour>=0.1.5',
    ]
)



# Find the set of maximums and minimums in a series,
# with some tolerance for multiple max or minimums 
# if some highest or lowest values in a series are
# tolerance_threshold away
def find_maxs_or_mins_in_series(series, kind, tolerance_threshold):
  isolated_globals = []
  if kind == "max":
    global_max_or_min = series.max()
  else:
    global_max_or_min = series.min()
  
  for idx, val in series.items():
    if val == global_max_or_min or abs(global_max_or_min - val ) < tolerance_threshold:
      isolated_globals.append(idx)
  return isolated_globals

# Find the average distance between High and Low price in a set of candles
def avg_candle_range(candles):
  return max((candles.df.High - candles.df.Low).mean(), 0.01)

# Find mean in a list of int or floats
def mean(ls):
  total = 0
  for n in ls: total+=n
  return total/len(ls)


import numpy as np
import pandas as pd
import sys

from . import util
from . import structs

DEFAULT_CONFIG = {
  # For some price at date t, what difference must be exceeded between p_t-1 and p_t+1
  # with respsect to p_t for t to be considered pivot
  "pivot_seperation_threshold": lambda candles: util.avg_candle_range(candles) * 0.2,

  # For some pivot found at date t, what difference is tolerable between p_t-1 to p_t and p_t+1 to p_t
  # for either p_t-1 or p_t+1 to ALSO be considered pivots
  "pivot_grouping_threshold": lambda candles: util.avg_candle_range(candles) * 0.1,

  # Max allowable error for a trendline point and a candlestick price
  "max_allowable_error_pt_to_trend": lambda candles: util.avg_candle_range(candles) * 0.06,
  
  # After trendlines have been found, the slopes and last prices for each trendline are collected in a 2D
  # array. We use 2D clustering of (slope,last_price) to find 'duplicates' ie almost identical trendlines
  # Below are the thresholds for grouping last price and slope
  "duplicate_grouping_threshold_last_price": lambda candles: util.avg_candle_range(candles) * 0.2,
  "duplicate_grouping_threshold_slope": lambda candles: util.avg_candle_range(candles) * 0.05,

  # How much does a trendline break into any candle for it to be considered a break-out
  "breakout_tolerance": lambda candles: util.avg_candle_range(candles) * 0.08,

  # Scores a detected trendline given a slice of distance from trend to price, slope, and candlesticks
  "scoring_function": lambda candles, err_distances, num_points, slope: (util.avg_candle_range(candles) / util.mean(err_distances)) * (2.5 ** num_points),

  # Max and min allowable slope angle for both resistance and support lines
  # By default set to allow all angles but min or max can be set to 0 to only allow possitive / negative slopes
  'max_allowable_support_slope': lambda candles: sys.float_info.max,
  'min_allowable_support_slope': lambda candles: -sys.float_info.max,
  'max_allowable_resistance_slope': lambda candles: sys.float_info.max,
  'min_allowable_resistance_slope': lambda candles: -sys.float_info.max,

  # Max and min allowable last price point for both resistance and support lines
  # By default set to be 1.5X of last candle closing price for max and 0.667x for min 
  'max_allowable_support_last_price': lambda candles: candles.df.Close.iloc[-1] * 1.5,
  'min_allowable_support_last_price': lambda candles: candles.df.Close.iloc[-1] * 0.667,
  'max_allowable_resistance_last_price': lambda candles: candles.df.Close.iloc[-1] * 1.5,
  'min_allowable_resistance_last_price': lambda candles: candles.df.Close.iloc[-1] * 0.667,
}

def get_pivots(
  candlestick_data=None,
  trend_type=None,
  scan_from_index=None,
  config=DEFAULT_CONFIG,
  debug=False
):
  if candlestick_data == None:
    raise Exception("No candlestick data provided")
  elif type(candlestick_data) != structs.CandlestickData:
    raise Exception("candlestick_data input provided is of invalid type. See README for instructions")

  if trend_type == None:
    raise Exception("No trend_type data provided")
  elif type(trend_type) != str:
    raise Exception("trend_type input provided is of invalid type. See README for instructions")

  separation_thres = config.get('pivot_seperation_threshold', DEFAULT_CONFIG['pivot_seperation_threshold'])(candlestick_data)
  grouping_thres = config.get('pivot_grouping_threshold', DEFAULT_CONFIG['pivot_grouping_threshold'])(candlestick_data)

  col = "Low" if trend_type == structs.TrendlineTypes.SUPPORT else "High"  
  pseries = candlestick_data.df[col][scan_from_index:]
  pivots = set([])
  max_number_continuous_pivots = 6

  first_index = pseries.index[0]
  last_index = pseries.index[-1]

  for i in range(first_index+1,last_index):
    pcur = pseries[i]
    pprev = pseries[i-1]

    j = 1
    while j < max_number_continuous_pivots and (i + j) < len(pseries) - 1:
      if abs(pseries[i + j - 1] - pseries[i + j]) < grouping_thres:
        j += 1
      else:
        break

    pnext = pseries[i + j]

    j = 1
    while j < max_number_continuous_pivots and (i - j) > first_index:
      if abs(pseries[i - j + 1] - pseries[i - j]) < grouping_thres:
        j += 1
      else:
        break
    
    pprev = pseries[i - j]

    if trend_type == structs.TrendlineTypes.RESISTANCE and \
      (pprev > pcur or pnext > pcur): continue
    elif trend_type == structs.TrendlineTypes.SUPPORT and \
      (pprev < pcur or pnext < pcur): continue

    if abs(pcur - pprev) > separation_thres * (1/4) and abs(pcur - pnext) > separation_thres * (3/4):
      pivots.add(i)
    elif abs(pcur - pprev) > separation_thres * (3/4) and abs(pcur - pnext) > separation_thres * (1/4):
      pivots.add(i)

  # Always include last point and first point as pivots
  pivots.add(first_index)
  pivots.add(last_index)

  if debug:
    print(
      "🐛 FIND PIVOT DEBUG 🐛\n" + \
      "Grouping Threshold = {}\n".format(grouping_thres) + \
      "Separation Threshold = {}\n".format(separation_thres) + \
      "Pivots found = {}\n".format(pivots) + \
      "Percentage of pivots to total data points = {}\n".format(((len(pivots) / len(pseries))) * 100 )
    )
    
  return pivots


def detect(
  candlestick_data=None,
  trend_type=None,

  # Specify if you require the first point of a trendline to be a pivot
  first_pt_must_be_pivot=False,
  # Specify if you require the last point of the trendline to be a pivot
	last_pt_must_be_pivot=False,
	# Specify if you require all trendline points to be pivots
  all_pts_must_be_pivots=False,
	# Specify if you require one of the trendline points to be global max or min price
  trendline_must_include_global_maxmin_pt=False,
  # Specify minimum amount of points required for trendline detection (NOTE: must be at least two)
  min_points_required=3,
  # Specify if you want to ignore prices before some date
  scan_from_date=None,
  # Specify if you want to ignore 'breakout' lines. That is, lines that interesect a candle
  ignore_breakouts=True,

  # Specify config for tuning detection/grouping/scoring parameters
  config=DEFAULT_CONFIG,

  # Console print debugs
  debug=False
):
  def detect_wrapped(tt):
    '''
    The algorithm will fly through all N^2 pivot point pairs,
    then it attempts to draw an aproximation line through each pair
    and will find this line to be a valid trendline iff this line contains
    points within config.max_allowable_error_pt_to_trend and and a total point
    quantity of min_points_required

    After all trendlines are retrieved:
      - Retrieve score
      - Detect which of the trendlines are duplicate (ie their slope and intercept are almost identical)
        and mark the one with best score

    '''
    # Input validation
    if candlestick_data == None:
      raise Exception("No candlestick data provided")
    elif type(candlestick_data) != structs.CandlestickData:
      raise Exception("candlestick_data input provided is of invalid type. See README for instructions")

    if tt == None:
      raise Exception("No trend_type data provided")
    elif type(tt) != str:
      raise Exception("trend_type input provided is of invalid type. See README for instructions")

    # Process config
    max_allowable_error_pt_to_trend = config.get("max_allowable_error_pt_to_trend", DEFAULT_CONFIG["max_allowable_error_pt_to_trend"])(candlestick_data)
    breakout_tolerance = config.get("breakout_tolerance", DEFAULT_CONFIG["breakout_tolerance"])(candlestick_data)
    max_allowable_support_slope = config.get("max_allowable_support_slope", DEFAULT_CONFIG["max_allowable_support_slope"])(candlestick_data)
    min_allowable_support_slope = config.get("min_allowable_support_slope", DEFAULT_CONFIG["min_allowable_support_slope"])(candlestick_data)
    max_allowable_resistance_slope = config.get("max_allowable_resistance_slope", DEFAULT_CONFIG["max_allowable_resistance_slope"])(candlestick_data)
    min_allowable_resistance_slope = config.get("min_allowable_resistance_slope", DEFAULT_CONFIG["min_allowable_resistance_slope"])(candlestick_data)
    max_allowable_support_last_price = config.get("max_allowable_support_last_price", DEFAULT_CONFIG["max_allowable_support_last_price"])(candlestick_data)
    min_allowable_support_last_price = config.get("min_allowable_support_last_price", DEFAULT_CONFIG["min_allowable_support_last_price"])(candlestick_data)
    max_allowable_resistance_last_price = config.get("max_allowable_resistance_last_price", DEFAULT_CONFIG["max_allowable_resistance_last_price"])(candlestick_data)
    min_allowable_resistance_last_price = config.get("min_allowable_resistance_last_price", DEFAULT_CONFIG["min_allowable_resistance_last_price"])(candlestick_data)

    max_allowable_slope = max_allowable_support_slope if tt == structs.TrendlineTypes.SUPPORT else max_allowable_resistance_slope
    min_allowable_slope = min_allowable_support_slope if tt == structs.TrendlineTypes.SUPPORT else min_allowable_resistance_slope
    max_allowable_last_price = max_allowable_support_last_price if tt == structs.TrendlineTypes.SUPPORT else max_allowable_resistance_last_price
    min_allowable_last_price = min_allowable_support_last_price if tt == structs.TrendlineTypes.SUPPORT else min_allowable_resistance_last_price

    if scan_from_date != None:
      scan_from_index = candlestick_data.df.loc[candlestick_data.df["Date"] == scan_from_date]
    else:
      scan_from_index = candlestick_data.df.loc[0].name

    col = "Low" if tt == structs.TrendlineTypes.SUPPORT else "High"  
    pseries = candlestick_data.df[col][scan_from_index:]
    pseries_sub = pseries[scan_from_index:]

    last_index = len(candlestick_data.df) - 1
    last_price = pseries[len(candlestick_data.df) - 1]
    time_interval_min = candlestick_data.time_interval_min()

    pivots = get_pivots(candlestick_data, tt, scan_from_index, config)
    pivots_sorted = list(pivots)
    pivots_sorted.sort()

    avg_candle_range = util.avg_candle_range(candlestick_data)

    # Threshold to decide max difference from a global max/min and consecutive best next global max/min to both be considered
    max_or_min_capture_thres = avg_candle_range * 0.10
    global_max_or_mins = []
    if tt == structs.TrendlineTypes.RESISTANCE:
      global_max_or_mins = util.find_maxs_or_mins_in_series(pseries_sub, "max", max_or_min_capture_thres)
    else:
      global_max_or_mins = util.find_maxs_or_mins_in_series(pseries_sub, "min", max_or_min_capture_thres)

    # Trendlines is a pandas dataframe containing columns ( slice_of_points, num_points, slope, intercept, score)
    trends_df = pd.DataFrame(columns=[
      'id',
      'trendtype',
      'pointset_indeces',
      'pointset_dates',
      'starts_at_index',
      'starts_at_date',
      'ends_at_index',
      'ends_at_date',
      'is_breakout',
      'breakout_index',
      'breakout_date',
      'num_points',
      'm',
      'b',
      'slope',
      'price_at_last_date',
      'score',
      'includes_global_max_or_min',
      'global_maxs_or_mins',
      'price_at_next_future_date',
      'duplicate_group_id',
      'is_best_from_duplicate_group',
      'overall_rank',
      'rank_within_group',
    ])
    
    for i in range(0, len(pseries)):
      # If we only specify using pivot points as start, skip non pivots
      if (first_pt_must_be_pivot or all_pts_must_be_pivots) and i not in pivots:
          continue

      # Skip indeces after opts.scan_from_index
      if scan_from_index > i:
        continue

      for j in range(i+1, len(pseries)):
        # If we only specify using pivot points as end, skip non pivots
        if (last_pt_must_be_pivot or all_pts_must_be_pivots) and j not in pivots:
            continue

        iprice = pseries[i]
        jprice = pseries[j]

        # Find the slope and intercept made formed by these two points
        m, b = np.polyfit([i, j], [iprice, jprice], 1)

        # Slope is found by considering time_interval_min as rightward unit and average candle range as upward unit
        slope = m * avg_candle_range
      
        # Check if the estimated price at last date or slope is allowable
        trend_price_at_last = m * last_index + b


        if slope > max_allowable_slope or slope < min_allowable_slope:
          continue

        if trend_price_at_last > max_allowable_last_price or trend_price_at_last < min_allowable_last_price:
          continue

        # Determine breakouts + count the number of points within this trendline
        num_points = 2
        points_in_trendline = [i, j]
        prices_in_trendline = [m * i + b, m * j + b]
        is_breakout = False
        breakout_index = None
        breakout_date = None

        for k in range(i, len(pseries)):
          if last_pt_must_be_pivot and k not in pivots:
            continue

          # Skip checking for the i or j case because these are already points in the set
          if k == i or k == j:
            continue

          trend_price_at_k = m * k + b

          # Determine if this trend is a breakout, if it hasn't been identified as one already
          if not is_breakout:
            if tt == structs.TrendlineTypes.RESISTANCE and trend_price_at_k < pseries[k] - breakout_tolerance or \
                tt == structs.TrendlineTypes.SUPPORT and trend_price_at_k > pseries[k] + breakout_tolerance:

              breakout_index = k
              breakout_date = candlestick_data.df.iloc[i].Date
              is_breakout = True
            

          if abs(trend_price_at_k - pseries[k]) < max_allowable_error_pt_to_trend:
            num_points += 1
            points_in_trendline.append(k)
            prices_in_trendline.append(trend_price_at_k)

        # Check if we have the minimum required number of points for trend
        if num_points < min_points_required: continue

        # Construct a "pointset_id" a unique identifier for this set of points
        points_in_trendline.sort()
        points_in_trendline_str = ('[' + str(points_in_trendline[0]))
        for p in range(1, len(points_in_trendline)):
          points_in_trendline_str += ("," + str(points_in_trendline[p]))
        points_in_trendline_str += "]"
        pointset_id = ("R" if tt == structs.TrendlineTypes.RESISTANCE else "S") + "-" + points_in_trendline_str

        # Determine if trendline has max or min, and skip this line if we require our lines have global max or min
        global_pt_found = False
        for p in points_in_trendline:
          if p in global_max_or_mins:
            global_pt_found = True

        if trendline_must_include_global_maxmin_pt and not global_pt_found:
          continue

        # We already have this pointset, just different order
        if pointset_id in trends_df.id.values: continue

        # We ignore this i,j pair if this is a breakout
        if is_breakout and ignore_breakouts: continue

        # Scoring
        err_distances = []
        for w in range(0,num_points):
          point_index = points_in_trendline[w]
          price_at_trendline = prices_in_trendline[w]
          price_actual = pseries[point_index]
          err_distances.append(abs(price_at_trendline - price_actual))
        
        score = config.get("scoring_function", DEFAULT_CONFIG["scoring_function"])(candlestick_data, err_distances, num_points, slope)
        
        trends_df = trends_df.append({
          'id':              pointset_id,
          'trendtype':       tt,
          'pointset_indeces':points_in_trendline,
          'pointset_dates':  [candlestick_data.df.iloc[pt].Date for pt in points_in_trendline],
          'starts_at_index': points_in_trendline[0],
          'starts_at_date':  candlestick_data.df.iloc[points_in_trendline[0]].Date,
          'ends_at_index':   points_in_trendline[-1],
          'ends_at_date':    candlestick_data.df.iloc[points_in_trendline[-1]].Date,
          'is_breakout':     is_breakout,
          'breakout_index':  breakout_index,
          'breakout_date':   breakout_date,
          'num_points':      num_points,
          'm': m,
          'b': b,
          'slope': slope,
          'price_at_last_date': trend_price_at_last,
          'score': score,
          'global_maxs_or_mins': global_max_or_mins,
          'includes_global_max_or_min': global_pt_found,
          'price_at_next_future_date': trend_price_at_last + m,
          'duplicate_group_id': None,
          'is_best_from_duplicate_group': False,
          'overall_rank': None,
          'rank_within_group': 0
        }, ignore_index=True)

    # Mark which of the trendlines are duplicate
    trends_df = _mark_duplicates(trends_df, candlestick_data, tt, config)
    trends_df = trends_df.sort_values(by='score', ascending=False)

    # Correct data types
    trends_df["is_breakout"] = trends_df["is_breakout"].astype(bool)

    return trends_df, pivots


  if trend_type == structs.TrendlineTypes.BOTH:
    support_trendlines, support_pivots = detect_wrapped(structs.TrendlineTypes.SUPPORT)
    resistance_trendlines, resistance_pivots = detect_wrapped(structs.TrendlineTypes.RESISTANCE)

    return {
      'trend_type': trend_type,
      'candlestick_data': candlestick_data,
      'support_pivots': support_pivots,
      'support_trendlines': support_trendlines,
      'resistance_pivots': resistance_pivots,
      'resistance_trendlines': resistance_trendlines,
    }
  elif trend_type == structs.TrendlineTypes.SUPPORT:
    support_trendlines, support_pivots = detect_wrapped(structs.TrendlineTypes.SUPPORT)

    return {
      'trend_type': trend_type,
      'candlestick_data': candlestick_data,
      'support_pivots': support_pivots,
      'support_trendlines': support_trendlines,
    }
  else:
    resistance_trendlines, resistance_pivots = detect_wrapped(structs.TrendlineTypes.RESISTANCE)

    return {
      'trend_type': trend_type,
      'candlestick_data': candlestick_data,
      'resistance_pivots': resistance_pivots,
      'resistance_trendlines': resistance_trendlines
    }

def _mark_duplicates(trends_df, candlestick_data, trend_type, config):
  duplicate_grouping_threshold_last_price = config.get("duplicate_grouping_threshold_last_price", DEFAULT_CONFIG['duplicate_grouping_threshold_last_price'])(candlestick_data)
  duplicate_grouping_threshold_slope = config.get("duplicate_grouping_threshold_slope", DEFAULT_CONFIG['duplicate_grouping_threshold_slope'])(candlestick_data)
  
  if len(trends_df) == 0: return trends_df

  if len(trends_df) == 1:
    trends_df.loc[0, 'duplicate_group_id'] = 1000
    trends_df.loc[0, 'is_best_from_duplicate_group'] = True
    trends_df.loc[0, 'overall_rank'] = 1
    trends_df.loc[0, 'rank_within_group'] = 1
    return trends_df

  group_idx = 1000 if trend_type == structs.TrendlineTypes.RESISTANCE else 2000

  # 2D clustering last price and slope to find closely related trendlines
  for i in range(0, len(trends_df)):
    best_matching_idx = -1
    row_i = trends_df.iloc[i]
    row_last_price = row_i['price_at_last_date']
    row_slope_angle = row_i['slope']
    i_is_breakout = row_i['is_breakout']

    for j in range(0, len(trends_df)):
      if i==j: continue

      row_j = trends_df.iloc[j]
      sec_row_last_price = row_j['price_at_last_date']
      sec_row_slope_angle = row_j['slope']
      j_is_breakout = row_j['is_breakout']

      price_diff = abs(row_last_price - sec_row_last_price)
      slope_angle_diff = abs(row_slope_angle - sec_row_slope_angle)

      # We determine that row i and row j are a good match for same group
      if ((price_diff < duplicate_grouping_threshold_last_price) and (slope_angle_diff < duplicate_grouping_threshold_slope)) \
          and (i_is_breakout == j_is_breakout):
        best_matching_idx = j
        break

    if(best_matching_idx != -1):
      this_row = row_i
      best_matching_row = trends_df.iloc[best_matching_idx]

      this_row_group = this_row['duplicate_group_id']
      best_matching_row_group = best_matching_row['duplicate_group_id']

      group_id_for_pair = best_matching_row_group if best_matching_row_group is not None else group_idx

      if best_matching_row_group == None:
        trends_df.loc[best_matching_row.name, 'duplicate_group_id'] = group_id_for_pair

      if this_row_group == None:
        trends_df.loc[this_row.name, 'duplicate_group_id'] = group_id_for_pair
      else:
        for k in range(0, len(trends_df)):
          row_k = trends_df.iloc[k]
          if row_k['duplicate_group_id'] == this_row_group:
            trends_df.loc[row_k.name, 'duplicate_group_id'] = group_id_for_pair

    else:
      trends_df.loc[row_i.name, 'duplicate_group_id'] = group_idx
    group_idx += 1
  
  # For all the duplicate groups found, mark best for each group
  if(len(trends_df) == 1):
    trends_df.loc[ trends_df.iloc[0].name , 'is_best_from_duplicate_group'] = True
    trends_df.loc[ trends_df.iloc[0].name , 'overall_rank'] = 1
    trends_df.loc[ trends_df.iloc[0].name, 'rank_within_group'] = 1
    return

  best_results = trends_df.sort_values(by='score', ascending=True).groupby(["duplicate_group_id"]).last().reset_index()
  overall_rank = 1

  for i in range(0, len(best_results)):
    row = best_results.iloc[i]
    trends_df.loc[ trends_df['id'] == row['id'] , 'is_best_from_duplicate_group'] = True
    trends_df.loc[ trends_df['id'] == row['id'] , 'overall_rank'] = overall_rank
    overall_rank += 1

  # Mark rank within each group
  dup_groups = trends_df.groupby(["duplicate_group_id"])
  for group_name, df_group in dup_groups:
    group_rank = 1

    df_group_sorted = df_group.sort_values(by='score', ascending=False)

    for row_index, row in df_group_sorted.iterrows():
      trends_df.loc[row_index, 'rank_within_group'] = int(group_rank)
      group_rank += 1

  return trends_df


import os
import pandas as pd
import numpy as np

from bokeh.resources import CDN
from bokeh.models.widgets import Div
from bokeh.plotting import figure
from bokeh.models import Label
from bokeh.embed import file_html

from datetime import timedelta
from colour import Color

from math import pi, tan

from . import structs

css_hack = '''
.dataframe {
	border: 1px solid grey;
	border-collapse: separate;
	border-spacing: 15px 5px;
	white-space: nowrap;
}

/* SECTION SHOWING LOG HISTORY */
details {
  background: white;
  border: 1px solid grey;
  color: #333333;
  border-radius: 4px;
  overflow-x: scroll;
  margin-top: 30px;
  margin-bottom: 30px;
  max-width: fit-content;
}

details summary {
  font-family: Courier;
  font-size: 16px;
  display: block;
  cursor: pointer;
  padding-right: 10px;
  padding-left: 10px;
  padding-top: 5px;
  padding-bottom: 5px;
  color: #2b2b2b;

}

details:not([open]) summary:hover,
details:not([open]) summary:focus {      
  background: #ffffff;
  color: #454545;
}

details[open] summary {
  border: 1px solid #003eff;
}

details main {
  padding: 0em;
  font-family: Courier;
  font-size: 14px;
}
'''

class TrendlineFigure():
  # point_set can either be a slice of date strings or a slice of indices

  def __init__(self, trend_type, result_row, plotting_prop_overrides = {}):
    self.type = trend_type

    self.id = result_row['id']
    self.pointset_dates = result_row['pointset_dates']
    self.breakout_index = result_row['breakout_index']
    self.is_breakout = result_row['is_breakout']
    self.score = result_row['score']
    self.includes_global_max_or_min = result_row['includes_global_max_or_min']
    self.global_maxs_or_mins = result_row['global_maxs_or_mins']
    self.is_best_from_duplicate_group = result_row['is_best_from_duplicate_group']
    self.plotting_prop_overrides = plotting_prop_overrides

  def plot_figure(self, p, candles_df, opts={}):    
    pointset_indeces = []
    for date in self.pointset_dates:
      pointset_indeces.append( candles_df.loc[candles_df['Date'] == date].index[0] )

    pt_set_y = []
    pt_set_x = pointset_indeces

    for pt in pt_set_x:
      if self.type == structs.TrendlineTypes.RESISTANCE:
        pt_set_y.append(candles_df.iloc[pt].High)
      elif self.type == structs.TrendlineTypes.SUPPORT:
        pt_set_y.append(candles_df.iloc[pt].Low)
      else:
        pt_set_y.append(candles_df.iloc[pt].Low)

    # Calculate slope and intersect using first point and last point
    m, b = np.polyfit([pt_set_x[0], pt_set_x[-1]], [pt_set_y[0], pt_set_y[-1]], 1)

    if self.is_breakout:
      last_date_index = self.breakout_index + 0.05
      last_date_trendline_price = m * last_date_index + b
      p.x([last_date_index], [last_date_trendline_price], line_width=3, size=10, color="red", alpha=0.8)
    else:
      last_date_index = len(candles_df) - 1

    tl_vals_at_x = pt_set_y
    tl_y_at_last_date = m * last_date_index + b

    color = self.get_trendline_plot_color()

    p.segment(
      x0=[pt_set_x[0]],
      y0=[tl_vals_at_x[0]],
      x1=[last_date_index],
      y1=[tl_y_at_last_date],
      color=color,
      line_width=self.get_trendline_plot_line_width(),
      line_dash=self.get_trendline_plot_line_style()
    )

    # Draw plot score label
    if self.is_best_from_duplicate_group:
      label_text =  "Score " + str(self.id) + " = " + str(round(self.score, 2))
      if self.is_breakout: label_text += " (breakout at {})".format(self.breakout_index)
      
      label_x_pos = last_date_index + 2
      label_y_pos = tl_y_at_last_date + (2 * m)

      label = Label(x=label_x_pos, y=label_y_pos,
                 text=label_text, render_mode='css',
                 border_line_color=color, border_line_alpha=0.8,
                 background_fill_color='white', background_fill_alpha=1.0)
      p.segment(
        x0=[last_date_index],
        y0=[tl_y_at_last_date],
        x1=[last_date_index + 2],
        y1=[tl_y_at_last_date + (m * 2)],
        color=color,
        line_width=self.get_trendline_plot_line_width(),
        line_dash=self.get_trendline_plot_line_style()
      )
      p.add_layout(label)

    # Mark points that make up trendline
    p.square(pt_set_x, tl_vals_at_x, size=12, color=color, alpha=0.5)

    # Mark global maxs or mins
    for date in self.global_maxs_or_mins:
      query = candles_df.loc[candles_df['Date'] == date]
      if len(query) == 0: continue
      index = query.iloc[0].name

      if self.type == structs.TrendlineTypes.RESISTANCE:
        price = candles_df.iloc[index].High
      elif self.type == structs.TrendlineTypes.SUPPORT:
        price = candles_df.iloc[index].Low
      p.circle(index, price, size=20, color="gold", alpha=0.3)

  def get_trendtype_string(self):
    if self.type == structs.TrendlineTypes.RESISTANCE:
      return "Resistance"
    elif self.type == structs.TrendlineTypes.SUPPORT:
      return "Support"
    else:
      return "Unknown"

  def get_trendline_plot_color(self):
    # If we have set an override for plot color, use this
    if "color" in self.plotting_prop_overrides:
      return self.plotting_prop_overrides["color"]

    orange = Color("darkorange")
    purple = Color("steelblue")
    grey = Color("slategrey")

    if self.type == structs.TrendlineTypes.SUPPORT:
      c = purple
    else:
      c = orange
    
    if self.is_breakout:
      # Set color to grey if we find that line has expired
      c = grey

    
    return c.hex_l

  def get_trendline_plot_line_width(self):
    if "line_width" in self.plotting_prop_overrides:
      return self.plotting_prop_overrides["line_width"]

    if self.is_breakout:
      return 1
    
    else:
      if self.is_best_from_duplicate_group:
        return 2
      else:
        return 1

  def get_trendline_plot_line_style(self):
    if "line_style" in self.plotting_prop_overrides:
      return self.plotting_prop_overrides["line_style"]

    if self.is_breakout:
      return "dotted"
    elif self.is_best_from_duplicate_group:
      return "dashed"
    else:
      return "dotted"

def _draw_bidirectional_ray(p, x, y, angle, color, width=2, dash="dashed"):
  p.segment(x0=x, x1=x, y0=0, y1=10000, line_color=color, line_dash=dash, line_width=width)

def _highlight_pivots(p, pivots_indexes, col, candles_df):
  # Highlight pivot points
  pivots_x_vals = []
  pivots_y_vals = []
  for pivot_idx in pivots_indexes:
    pivot_panda_ts = candles_df.iloc[pivot_idx].name
    pivot_y_val = candles_df.iloc[pivot_idx][col]
    pivots_x_vals.append(pivot_panda_ts)
    pivots_y_vals.append(pivot_y_val)
  p.diamond(pivots_x_vals, pivots_y_vals, size=20, line_color="green", fill_alpha=0.1, alpha=0.5)

def plot_graph_bokeh(results):
  candlestick_data = results["candlestick_data"]

  # Plot
  candles_df = candlestick_data.df

  site_title = '{} trendlines for {}m candlesticks'.format(
    "RESISTANCE and SUPPORT" if results["trend_type"] == "BOTH" else results["trend_type"],
    candlestick_data.time_interval_min(),
  )
  
  
  # Establish plot dimensions
  y_padding = (candles_df['High'].max() / 200)
  y_range_top = candles_df['Low'].min() - y_padding
  y_range_bottom = candles_df['High'].max() + y_padding
  x_range_left = -1
  x_range_right = len(candles_df) + 10
  
  
  # Plot candlestick chart
  inc = candles_df.Close > candles_df.Open
  dec = candles_df.Open > candles_df.Close
  w = 0.5

  p = figure(
    x_axis_type="datetime",
    tools="pan,wheel_zoom,tap,crosshair,hover,poly_draw,reset,save",
    plot_width=1300,
    title=site_title,
    y_range=(y_range_top, y_range_bottom),
    x_range=(x_range_left, x_range_right),
  )
  
  p.xaxis.major_label_overrides = {
    i: date.strftime('%b %d %H:%M') for i, date in enumerate(candles_df['Date'])
  }

  p.xaxis.major_label_orientation = pi/4
  p.grid.grid_line_alpha=0.3
  p.segment(candles_df.index, candles_df.High, candles_df.index, candles_df.Low, color="black")
  p.vbar(candles_df.index[inc], w, candles_df.Open[inc], candles_df.Close[inc], fill_color="#D5E1DD", line_color="black")
  p.vbar(candles_df.index[dec], w, candles_df.Open[dec], candles_df.Close[dec], fill_color="#F2583E", line_color="black")

  # Plot trendlines (support)
  if 'support_trendlines' in results:
    for _, result_row in results['support_trendlines'].iterrows():
      tf = TrendlineFigure(structs.TrendlineTypes.SUPPORT, result_row)
      tf.plot_figure(p, candles_df)

  # Plot trendlines (resistsance)
  if 'resistance_trendlines' in results:
    for _, result_row in results['resistance_trendlines'].iterrows():
      tf = TrendlineFigure(structs.TrendlineTypes.RESISTANCE, result_row)
      tf.plot_figure(p, candles_df)

  # Draw vertical lines at first and last price
  _draw_bidirectional_ray(p, candles_df.index[0] - 0.5, 0, 90, "#bbbbbb")
  _draw_bidirectional_ray(p, candles_df.index[-1] + 0.5, 0, 90, "#bbbbbb")


  # Highlight pivot points
  if 'support_pivots' in results:
    _highlight_pivots(p, results['support_pivots'], "Low", candles_df)
  if 'resistance_pivots' in results:
    _highlight_pivots(p, results['resistance_pivots'], "High", candles_df)

  # Styling nits
  p.title.text_font_size = '16pt'

  return p

def plot_table_bokeh(results):
  if results['trend_type'] == structs.TrendlineTypes.BOTH:
    all_results = pd.concat([results['support_trendlines'], results['resistance_trendlines']])
  elif results['trend_type'] == structs.TrendlineTypes.SUPPORT:
    all_results = results['support_trendlines']
  else:
    all_results = results['resistance_trendlines']

  if len(all_results) > 0:
    html_trends_table = all_results.drop('pointset_dates', 1).to_html(border=0, header=True, index=False, justify="left", float_format=lambda x: '%10.3f' % x)
  else:
    html_trends_table = '<p>No trendlines found</p>'

  div = Div(text="", width=1000)
  div.text = '''
    <div>
      <h1>Trendline results:</h1><br/>
      {}
    </div>
  '''.format(html_trends_table)

  return div

def plot(
  results=None,
  filedir='.',
  filename='trend_plot.html',
):
  # Validate data
  if results == None or type(results) != dict:
    raise Exception("results argument for plot needs to be output of detect(...)")

  # Plot candlestick graph with trendlines
  trend_graph = plot_graph_bokeh(results)

  # Plot trendline results dataframe below the plot
  trend_table = plot_table_bokeh(results)

  # Get HTML content
  filepath = filedir + '/' + filename
  html_content = file_html(models=(trend_graph, trend_table), resources=(CDN), title="pytrendline results")

  # Hack in extra styles to HTML
  html_content = html_content.replace('<head>', '<head><style class="custom" type="text/css">{}</style>'.format(css_hack))

  # Write HTML to file
  with open(filepath, 'w') as outfile:
    outfile.write(html_content)
  
  return filepath


import pandas as pd
from pandas.api.types import is_datetime64_any_dtype as is_datetime
import datetime

VALID_TIME_INTERVALS = ["1m","3m","5m","10m","15m","30m","1h","1d"]

class TrendlineTypes(object):
  RESISTANCE = 'RESISTANCE'
  SUPPORT = 'SUPPORT'
  BOTH = 'BOTH'

class CandlestickData():
  def __init__(
    self,
    df=None,
    time_interval="1m", # choose between 1m,3m,5m,10m,15m,30m,1h,1d
    open_col="Open", # name of the column containing candle "Open" price
    high_col="High", # name of the column containing candle "High" price
    low_col="Low", # name of the column containing candle "Low" price
    close_col="Close", # name of the column containing candle "Close" price
    datetime_col=None # name of the column containing candle datetime price (use none if datetime is in index)
  ):
    # Validate dataframe nonempty and has at least 3 entries
    if df is None or len(df)<3:
      raise Exception("CandlestickData constructor param df should have at least three entries, received :\n{}".format(
        df
      ))

    # Validate time interval supplied is correct
    if time_interval not in VALID_TIME_INTERVALS:
      raise Exception("CandlestickData constructor param time_interval must be one of :\n{}".format(
        VALID_TIME_INTERVALS
      ))

    # Price column names provided exist and are in correct format
    for col_name in [open_col, high_col, low_col, close_col]:
      if col_name not in df.columns:
        raise Exception("CandlestickData constructor param df does not contain column '{}'".format(
          col_name
        ))
      if df[col_name].dtypes != float:
        raise Exception("CandlestickData constructor param df requires that column '{}' is of type float".format(
          col_name
        ))

    # Datetime column provide exists and is of type datetime (or index is of type datetime if arg is None)
    if datetime_col == None:
      if not is_datetime(df.index): 
        raise Exception(
          "If no datetime_col is provided to CandlestickData constructor, the df index must be of datetime type.\n" + 
          "Instead received df index of {} type".format(str(type(df.index)))
        )
    else:
      if datetime_col not in df.columns:
        raise Exception("CandlestickData constructor param df does not contain datetime_col '{}'".format(
          datetime_col
        ))
      if not is_datetime(df[datetime_col]):
        raise Exception("CandlestickData constructor param df requires that column '{}' be of type datetime, received {}".format(
          datetime_col, df[datetime_col].dtypes
        ))

    # Instantiate
    self.df = df.copy(deep=True)

    self.df = self.df.rename(columns={
      open_col: "Open",
      high_col: "High",
      low_col: "Low",
      close_col: "Close",
    })

    # Force index to be monotinically increasing int
    if datetime_col == None:
      self.df.set_index(datetime_col)
      self.df = self.df.rename(columns = {'index':'Date'})
    else:
      self.df.rename(columns={datetime_col: "Date"})
    
    self.df.reset_index()

    self.time_interval = time_interval
    self.open_col = open_col
    self.high_col = high_col
    self.low_col = low_col
    self.close_col = close_col
    self.datetime_col = datetime_col

  def time_interval_min(self):
    if 'm' in self.time_interval:
      return int(self.time_interval[:-1])
    elif 'h' in self.time_interval:
      return int(self.time_interval[:-1] * 60)
    elif 'd' in self.time_interval:
      return int(self.time_interval[:-1] * 60 * 24)
